@@ -1,8 +1,8 @@
-/* Time-stamp: <2005-12-12 10:10:58 poser>
+/* Time-stamp: <2006-04-17 19:06:02 poser>
  *
  * Convert text containing various 7-bit ASCII escapes to UTF-7 Unicode.
  *
- * Copyright (C) 2004, 2005 William J. Poser (billposer@alum.mit.edu)
+ * Copyright (C) 2005-2006 William J. Poser (billposer@alum.mit.edu)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -56,7 +56,7 @@ ShowVersion(void){
 #ifdef HAVE_DATE_TIME
   fprintf(stderr,"%s\n",compdate);
 #endif
-  fprintf(stderr,"Copyright (C) 2005 William J. Poser\n");
+  fprintf(stderr,"Copyright (C) 2005-2006 William J. Poser\n");
   fprintf(stderr,_("Released under the terms of the GNU General Public License.\n\n"));
 }
 
@@ -77,7 +77,7 @@ containing various representations for non-ASCII characters\nto UTF-8 Unicode.\n
   fprintf(stderr,
 	  _("       -C Convert \\x-escaped hexadecimal numbers in braces (\\x{00E9})\n"));
   fprintf(stderr,
-	  _("       -D Convert decimal numeric character references (&#0233;)\n"));
+	  _("       -D Convert decimal HTML numeric character references (&#0233;)\n"));
   fprintf(stderr,
 	  _("       -E Convert hexadecimal with prefix U (U00E9)\n"));
   fprintf(stderr,
@@ -85,23 +85,35 @@ containing various representations for non-ASCII characters\nto UTF-8 Unicode.\n
   fprintf(stderr,
 	  _("       -G Convert hexadecimal in single quotes with prefix X (X\'00E9\')\n"));
   fprintf(stderr,
-	  _("       -H Convert hexadecimal numeric character references (&#x00E9;)\n"));
+	  _("       -H Convert hexadecimal HTML numeric character references (&#x00E9;)\n"));
   fprintf(stderr,
-	  _("       -I Convert hexadecimal UTF-8 with each byte's hex preceded by an =-sign (=C3=A9)\n"));
+	  _("       -I Convert hexadecimal UTF-8 with each byte's hex preceded by an =-sign (=C3=A9)\n\t\tThis is the URI escape format defined by RFC 2396.\n"));
   fprintf(stderr,
-	  _("       -J Convert hexadecimal UTF-8 with each byte's hex preceded by a %%-sign  (%%C3%%A9)\n"));
+	  _("       -J Convert hexadecimal UTF-8 with each byte's hex preceded by a %%-sign  (%%C3%%A9).\n\t\tThis is the Quoted Printable format defined by RFC 2045.\n"));
   fprintf(stderr,
 	  _("       -K Convert octal UTF-8 with each byte escaped by a backslash (\\303\\251)\n"));
   fprintf(stderr,
-	  _("       -L Convert \\u-escaped hex (\\u00E9) within the BMP,\n\t\t\\U-escaped hex (\\U00010024) outisde it.\n"));
+	  _("       -L Convert \\u-escaped hex (\\u00E9) within the BMP (U+0000-U+FFFF),\n\t\t\\U-escaped hex (\\U00010024) outisde it.\n"));
+  fprintf(stderr,
+	  _("       -M Convert hexadecimal SGML numeric character references (\\#x00E9;)\n"));
+  fprintf(stderr,
+	  _("       -N Convert decimal SGML numeric character references (\\#0233;)\n"));
+  fprintf(stderr,
+	  _("       -O Convert octal escapes for the three low bytes in big-endian order (\\000\\000\\351)\n"));
   fprintf(stderr,
 	  _("       -P Convert hexadecimal numbers with prefix U+ (U+00E9)\n"));
   fprintf(stderr,
 	  _("       -Q Convert HTML character entities (&eacute;)\n"));
   fprintf(stderr,
 	  _("       -R Convert raw hexadecimal numbers (00E9)\n"));
+  fprintf(stderr,
+	  _("       -S Convert hexadecimal escapes for the three low bytes in big-endian order (\\x00\\x00\\xE9)\n"));
+  fprintf(stderr,
+	  _("       -T Convert decimal escapes for the three low bytes in big-endian order (\\d000\\d000\\d233)\n"));
   fprintf(stderr,
 	  _("       -U Convert \\u-escaped hex (\\u00E9)\n"));
+  fprintf(stderr,
+	  _("       -V Convert \\u-escaped decimal (\\u0233)\n"));
   fprintf(stderr,
 	  _("       -X Convert standard form hexadecimal numbers (0x00E9)\n"));
   fprintf(stderr,
@@ -132,14 +144,20 @@ int main (int ac, char *av[])
   char *Ifmt = "=%2lX"; 		/* UTF-8 */
   char *Jfmt = "%%%2lX"; 		/* UTF-8 */
   char *Kfmt = "\\%3lo"; 		/* UTF-8 */
-  char *Pfmt  = "U+%lX";
+  char *Mfmt = "\\#x%lX;"; 
+  char *Nfmt = "\\#%ld;";
+  char *Ofmt = "\\%03o\\%03o\\%03o";
+  char *Pfmt = "U+%lX";
   char *Qfmt = "&%[abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789];";
-  char *Rfmt  = "%lX"; 
+  char *Rfmt = "%lX"; 
+  char *Sfmt = "\\x%02x\\x%02x\\x%02x";	
+  char *Tfmt = "\\d%03d\\d%03d\\d%03d";
   char *Ufmt = "\\u%8lX";
-  char *Xfmt  = "0x%lX";
+  char *Vfmt = "\\u%8ld";
+  char *Xfmt = "0x%lX";
 
   char *fmt = Xfmt;		/* Default is plain hex format */
-  char afmt [67+2+1];
+  char afmt [67+2+1+2];
   char aHfmt [8+2+1];
   char aDfmt [8+2+1];
   char cbuf[5];
@@ -162,6 +180,10 @@ int main (int ac, char *av[])
   int eof;
   char SplitStr[3];
   char enam[LBUFSIZE];
+  long MicrosoftStyle = 0L;
+  unsigned char b1;		/* Used for byte-wise encoding */
+  unsigned char b2;
+  unsigned char b3;
 
   extern int optind;
   extern int opterr;
@@ -181,7 +203,7 @@ int main (int ac, char *av[])
 
   /* Handle command line arguments */
 
-   while( (oc = getopt(ac,av,"ABCDEFHIJKLPQRUXYZ:hpqv")) != EOF){
+   while( (oc = getopt(ac,av,"ABCDEFHIJKLMNOPQRSTUVXYZ:hpqv")) != EOF){
      switch(oc){
      case 'A':
        fmt = Afmt;
@@ -189,6 +211,15 @@ int main (int ac, char *av[])
      case 'X':
        fmt = Xfmt;
        break;
+     case 'O':
+       fmt = Ofmt;
+       break;
+     case 'S':
+       fmt = Sfmt;
+       break;
+     case 'T':
+       fmt = Tfmt;
+       break;
      case 'P':
        fmt = Pfmt;
        break;
@@ -238,9 +269,18 @@ int main (int ac, char *av[])
        fmt = SplitFormat;
        BMPSplit =1;
        break;
+     case 'M':
+       fmt = Mfmt;
+       break;
+     case 'N':
+       fmt = Nfmt;
+       break;
      case 'U':
        fmt = Ufmt;
        break;
+     case 'V':
+       fmt = Vfmt;
+       break;
      case 'Y':
        fmt = Qfmt;
        AllHTMLP = 1;
@@ -262,6 +302,9 @@ int main (int ac, char *av[])
        ShowVersion();
        exit(INFO);
        break; 			/* NOTREACHED */
+     case ':':
+       fprintf(stderr,_("%s: missing argument to option flag %c.\n"),pgname,optopt);
+       exit(BADOPTIONARG);
      default:
        fprintf(stderr,_("%1$s: invalid option flag %2$c\n"),pgname,optopt);
        ShowVersion();
@@ -275,11 +318,17 @@ int main (int ac, char *av[])
      exit(BADOPTION);
    }
 
+   if(AllHTMLP && PureP) {
+     fprintf(stderr,_("Conversion of all three HTMl formats is not supported in pure mode.\n"));
+     exit(BADOPTION);
+   }
+
    if(AllHTMLP) {
      sprintf(aDfmt,"%s%%n",Dfmt);
      sprintf(aHfmt,"%s%%n",Hfmt);
    }
 
+   sprintf(afmt,"%s%%n",fmt);	/* Add %n for NConsumed */
    ReplacedNumber = 0L;
    TokenNumber = 0L;
    /*
@@ -298,26 +347,64 @@ int main (int ac, char *av[])
 		 pgname,TokenNumber);
 	 exit(OUTOFMEMORY);
        }
-       
-       Converted = sscanf(str,fmt,&num);
+       if(fmt == Qfmt) {
+	 Converted = sscanf(str,afmt,&enam,&NConsumed);
+	 num = LookupCodeForEntity(enam);
+	 if(!num) {
+	   num = UNI_REPLACEMENT_CHAR;
+	   fprintf(stderr,"ascii2uni: unknown HTML character entity \"&%s;\"\n",
+		   enam);
+	   ReplacedNumber++;
+	   Converted = (-1);
+	 }
+	 else Converted = 1;
+       }
+       else if( (Ofmt == fmt) || (Sfmt == fmt) || (Tfmt == fmt)) {
+	 Converted = sscanf(str,afmt,&b1,&b2,&b3,&NConsumed);
+	 switch(Converted)
+	   {
+	   case 3:
+	     num = (((b1 * 256) + b2) * 256) + b3;
+	     break;
+	   case 2:
+	     num = (b1 * 256) + b2;
+	     break;
+	   case 1:
+	     num = b1;
+	     break;
+	   default:
+	     break;
+	     /* This case is handled below */
+	 }
+       }
+       else {
+	 Converted = sscanf(str,afmt,&num,&NConsumed);
+       }
+
        if(Converted < 1) {
 	 fprintf(stderr,_("Ill-formed input %1$s at token %2$lu\n"),str,TokenNumber);
 	 exit(BADRECORD); 
        }
+       else if(Converted > 3) {
+	 fprintf(stderr,_("The character encoded as %1$s at token %2$lu is outside the Unicode range.\n\tEmitting Unicode replacement character.\n"),
+		 str,TokenNumber);
+	 putu8(UNI_REPLACEMENT_CHAR);
+       } 
        else {
 	 if (UTF8ValueP) putchar(num);
 	 else putu8(num);
+	 if( (fmt == Dfmt) || (fmt == Hfmt) || (fmt == Qfmt)) {
+	   if(*(str+NConsumed-1) != ';') MicrosoftStyle++;
+	 }
        }
        free((void *)str);
      }
-     if(VerboseP) fprintf(stderr,_("%ld tokens converted\n"),TokenNumber);
-     exit(SUCCESS);
+     goto done;
    }
 
    /* This is the case in which the Unicode escapes are embedded in ASCII text */
 
    LineNo = 0;
-   sprintf(afmt,"%s%%n",fmt);
    while(fgets(lbuf,LBUFSIZE,stdin) != NULL) {
      LineNo++;
      iptr = lbuf;
@@ -392,12 +479,14 @@ int main (int ac, char *av[])
 	   if(sscanf(iptr,aHfmt,&num,&NConsumed)) {
 	     putu8(num);
 	     iptr+=NConsumed;
+	     if(*(iptr-1) != ';') MicrosoftStyle++;
 	     TokenNumber++;
 	     continue;
 	   }
 	   else if(sscanf(iptr,aDfmt,&num,&NConsumed)) {
 	     putu8(num);
 	     iptr+=NConsumed;
+	     if(*(iptr-1) != ';') MicrosoftStyle++;
 	     TokenNumber++;
 	     continue;
 	   }
@@ -406,6 +495,7 @@ int main (int ac, char *av[])
 	   if( (num = LookupCodeForEntity(enam))) {
 	     putu8(num);
 	     iptr+=NConsumed;
+	     if(*(iptr-1) != ';') MicrosoftStyle++;
 	     TokenNumber++;
 	   }
 	   else {
@@ -417,21 +507,59 @@ int main (int ac, char *av[])
 	   }
 	 }
 	 else putchar(*iptr++);
+       } /* End of Qfmt case */
+       else if( (Ofmt == fmt) || (Sfmt == fmt) || (Tfmt == fmt)) {
+	 Converted=sscanf(iptr,afmt,&b1,&b2,&b3,&NConsumed);
+	 switch(Converted)
+	   {
+	   case 3:
+	     num = (((b1 * 256) + b2) * 256) + b3;
+	     putu8(num);iptr+=NConsumed;
+	     break;
+	   case 2:
+	     num = (b1 * 256) + b2;
+	     putu8(num);iptr+=NConsumed;
+	     break;
+	   case 1:
+	     num = b1;
+	     putu8(num);iptr+=NConsumed;
+	     break;
+	   case 0:
+	     putchar(*iptr++);
+	     break;
+	   default:
+	     fprintf(stderr,_("The character encoded as %1$s at token %2$lu is outside the Unicode range.\n\tEmitting Unicode replacement character.\n"),
+		     str,TokenNumber);
+	     putu8(UNI_REPLACEMENT_CHAR);
+	   }
+	   TokenNumber++;
        }
-       else {			/* Defaut - not BMPSplit and not HTML */
+       else {			/* Default - not BMPSplit, HTML, or byte format */
 	  if(sscanf(iptr,afmt,&num,&NConsumed)) {
 	   if (UTF8ValueP) putchar(num);
 	   else putu8(num);
 	   iptr+=NConsumed;
+	   if(fmt == Hfmt) {
+	     if(*(iptr-1) != ';') MicrosoftStyle++;
+	   }
+	   else if(fmt == Dfmt) {
+	     if(*(iptr-1) != ';') MicrosoftStyle++;
+	   }
 	   TokenNumber++;
 	 }
 	 else putchar(*iptr++);
        }
      } /* Loop over current line */
    } /* Loop over input lines */
+
+done:
    if(VerboseP) {
      fprintf(stderr,_("%ld tokens converted\n"),TokenNumber);
-     fprintf(stderr,_("%ld tokens replaced\n"),ReplacedNumber);
+     fprintf(stderr,_("%ld tokens replaced with Unicode Replacement Character\n"),ReplacedNumber);
+     if(MicrosoftStyle) {
+       fprintf(stderr,
+	       _("%ld Microsoft-style (lacking final semi-colon)\n"),MicrosoftStyle);
+     }
    }
    exit(SUCCESS);
 }
